import sys
import re
from inspect import signature, Parameter
from json import loads
from typing import Callable, Union, List, Tuple, Dict, Set, Any, Optional, get_args, get_origin
from types import MappingProxyType
from ast import literal_eval
from .output import prt, NbCmdIO

re_int = re.compile(r"^-?\d+$")
re_float = re.compile(r"^-?\d+\.\d+$")
re_keyword = re.compile(r"^([a-zA-Z_]\w{0,12})=(.*)$")  # 默认参数名不超过13个字符


# 可用函数字典
FUNCTIONS: Dict[object, list[str]] = {
    prt: [
        "setTitle",
        "cls",
        "print",
        "reset",
        "end",
        "loc",
        "col",
        "getLoc",
        "hideCursor",
        "showCursor",
        "bold",
        "dim",
        "italics",
        "underline",
        "blink",
        "invert",
        "strike",
        "fg_black",
        "bg_black",
        "fg_red",
        "bg_red",
        "fg_green",
        "bg_green",
        "fg_yellow",
        "bg_yellow",
        "fg_blue",
        "bg_blue",
        "fg_magenta",
        "bg_magenta",
        "fg_cyan",
        "bg_cyan",
        "fg_grey",
        "bg_grey",
        "fg_rgb",
        "bg_rgb",
        "fg_hex",
        "bg_hex",
        "getSize",
        "alignCenter",
        "alignRight",
        "setOrigin",
        "setOriginTerm",
        "drawNL",
        "drawHLine",
        "drawVLine",
        "drawRect",
        "printLines",
        "drawHGrad",
        "drawVGrad",
        "drawImage",
        "drawImageStr",
        "playGif",
        "playVideo",
        "clearRegion",
        "test",
        "fun",
    ]
}
def param_test(a, i:int, s:str, t:tuple, u:Union[list,tuple,str], *args, n=None, l=[], f=0.1, d={"k":1}, **kwds):
    print(f"a:{a}, i:{i}, s:{s}, t:{t}, u:{u}, args:{args}, n:{n}, l:{l}, f:{f}, d:{d}, kwds:{kwds}")

setattr(prt, 'fun', param_test)

AllFun = []


def get_fun(func: str) -> Callable:
    for obj in FUNCTIONS:
        if func in FUNCTIONS[obj]:
            return getattr(obj, func)
    return lambda *_: None


def list_functions():
    prt.bold().fg_hex("cff")("Functions:\n\n")
    sty_name = prt.bold().fg_hex("6f6").makeStyle()
    sty_args = prt.reset().fg_hex("ccf").makeStyle()
    for obj in FUNCTIONS:
        funs = "\n".join(
            [
                f"    {sty_name}{i}{sty_args}{str(signature(get_fun(i)))}\n"
                for i in FUNCTIONS[obj]
            ]
        )
        prt(funs).drawNL()
    prt("For detailed information about function, type:\n\n")
    prt.bold().fg_yellow("    help <function>\n").end()


def help_function(func=None):
    if func == None:
        yellow = prt.fg_yellow().makeStyle()
        prt.bold().drawHGrad(
            (220, 110, 80),
            (80, 150, 230),
            string=f"  NbCmdIO({prt.__version__}) prt cli.  ",
        ).drawNL()
        prt.bold("Usage")(": prt func1 args func2 args...\n")
        prt("Example:\n\n")
        prt.col(4).fg_blue(f"prt {yellow}loc{prt.RESET} 3 4 {yellow}drawImage{prt.RESET} filepath\n")
        prt.col(4).fg_blue(f'prt {yellow}fg_hex{prt.RESET} "#ccf" {yellow}bold{prt.RESET} "text"\n\n')
        prt.bold().fg_yellow("prt list")(": list all available functions\n")
        prt.bold().fg_yellow("prt help <function>")(
            ": get help information of function\n"
        ).end()
        return
    fun = get_fun(func)
    prt("Function ").bold().fg_yellow(func)(":\n")
    prt.col(5).bold().fg_hex("6f6")(func)(str(signature(fun))).drawNL()
    prt.col(7)(fun.__doc__).drawNL().end()


def parse_type(s: str, target_type: Any) -> Any:
    if s is None or s == 'null':
        return None
    origin = get_origin(target_type)
    args = get_args(target_type)
    # Optional (Union[T, None])
    if origin is Union and type(None) in args:
        actual_type = [t for t in args if t is not type(None)][0]
        return parse_type(s, actual_type)
    
    # Union 
    if origin is Union:
        for possible_type in args:
            try:
                return parse_type(s, possible_type)
            except (ValueError, TypeError):
                continue
        raise ValueError(f"Can not parse '{s}' as {target_type}.")
    
    # List
    if origin is list or target_type is list:
        item_type = args[0] if args else Any
        try:
            lst = literal_eval(s)
            if not isinstance(lst, list):
                raise ValueError(f"'{s}' is not a valid list.")
            return [parse_type(str(item), item_type) for item in lst]
        except (ValueError, SyntaxError):
            raise ValueError(f"Can not parse '{s}' as list")
    
    # Tuple
    if origin is tuple or target_type is tuple:
        try:
            tpl = literal_eval(s)
            if not isinstance(tpl, tuple):
                raise ValueError(f"'{s}' is not a valid tuple.")
            type_args = args
            if not type_args:
                return tpl
            if len(type_args) == 1 and type_args[0] == ():  # empty tuple
                return ()
            if len(tpl) != len(type_args):
                raise ValueError(f"Tuple length not match, expected {len(type_args)} but got {len(tpl)}.")
                
            return tuple(parse_type(str(item), type_args[i]) for i, item in enumerate(tpl))
        except (ValueError, SyntaxError):
            raise ValueError(f"Can not parse '{s}' as tuple.")
    
    # Dict
    if origin is dict or target_type is dict:
        try:
            d = literal_eval(s)
            if not isinstance(d, dict):
                raise ValueError(f"'{s}' is not a valid dict.")
            type_args = args
            if not type_args:  # normal dict
                return d
            key_type, value_type = type_args
            return {
                parse_type(str(k), key_type): parse_type(str(v), value_type)
                for k, v in d.items()
            }
        except (ValueError, SyntaxError):
            raise ValueError(f"Can not parse '{s}' as dict.")
    
    # Set
    if origin is set or target_type is set:
        try:
            item_type = args[0] if args else Any
            set_value = literal_eval(s)
            if not isinstance(set_value, (set, list, tuple)):
                raise ValueError(f"'{s}' is not a valid set.")
            return {parse_type(str(item), item_type) for item in set_value}
        except (ValueError, SyntaxError):
            raise ValueError(f"Can not parse '{s}' as set.")
    
    # base type
    if target_type is str:
        return s
        
    if target_type is int:
        try:
            return int(s)
        except ValueError:
            raise ValueError(f"Can not parse '{s}' as int.")
    
    if target_type is float:
        try:
            return float(s)
        except ValueError:
            raise ValueError(f"Can not parse '{s}' as float.")
    
    if target_type is bool:
        s_lower = s.lower()
        if s_lower in ('true', 't', 'yes', 'y', '1'):
            return True
        elif s_lower in ('false', 'f', 'no', 'n', '0'):
            return False
        raise ValueError(f"Can not parse '{s}' as bool.")
    
    # (Literal)
    if hasattr(target_type, '__args__'):
        if s in target_type.__args__:
            return s
        raise ValueError(f"'{s}' is not valid {target_type}.")
    
    return s

def get_param_type(param:Parameter) -> Any:
    if param.kind == param.VAR_POSITIONAL:
        return param.VAR_POSITIONAL
    if param.kind == param.VAR_KEYWORD:
        return param.VAR_KEYWORD
    if param.annotation != param.empty:
        return param.annotation
    if param.default is not param.empty:
        return type(param.default)
    return Any

def parse_args(args: List[str], params:MappingProxyType):
    """解析参数列表为参数字典
    目前*args和**kwds均解析为字符串，是否依据模式自定义解析有待商榷"""
    POS_ANY, KEY_ANY, KEY = False, False, False
    pos_args, key_args = [], {}
    pi = 0
    keys = list(params.keys())
    kvs = list(params.items())
    tps = [get_param_type(t) for _, t in kvs]
    has_kwds = Parameter.VAR_KEYWORD in tps
    for arg in args:
        m = re_keyword.match(arg)
        if KEY or m:
            KEY = True
            if not m:
                raise ValueError(f"Positional parameter after Keywords parameter: {arg}")
            key, value = m[1], m[2]
            if KEY_ANY:
                key_args[key] = parse_type(value, Any)
                continue
            if key in keys:
                tp = get_param_type(params[key])
                if tp == 0xe: 
                    KEY_ANY, tp = True, Any
                key_args[key] = parse_type(value, tp)
            elif has_kwds:
                key_args[key] = parse_type(value, Any)
            else:
                raise NameError(f"Unknown parameter keyword: {key}")
            continue
        if POS_ANY: tp = Any
        else: 
            tp = tps[pi]
            pi += 1
            if tp == Parameter.VAR_POSITIONAL: 
                POS_ANY, tp = True, Any
        pos_args.append(parse_type(arg, tp))
    return pos_args, key_args


def call_function(func_name: str, args: List[str]) -> None:
    """调用指定函数"""
    func = get_fun(func_name)
    sig = signature(func)
    params = sig.parameters

    pos_args, key_args = parse_args(args, params)
    try:
        res = func(*pos_args, **key_args)
    except KeyboardInterrupt:
        sys.exit(0)
    if func_name == "getSize":
        prt(str(res))


def parse_cmd_argv(argv: list[str]):
    l = len(argv)
    fun = argv[0]
    if fun not in AllFun:
        prt("Unknown function: ").fg_red(fun)
        return []
    args, funs = [], []
    for i in range(1, l):
        if argv[i] in AllFun:
            funs.append((fun, args))
            fun = argv[i]
            args = []
        else:
            args.append(argv[i])
    funs.append((fun, args))
    return funs


def cli(argv=[]):
    """主函数，处理命令行参数"""
    argv = sys.argv[1:] or argv
    if not argv:
        row, col = prt.getSize()
        if row > 13 and col > 42:
            NbCmdIO()
        help_function()
        return
    
    if not sys.stdin.isatty():  # 如果stdin不是终端(说明有管道输入)
        data = sys.stdin.read()
        if isinstance(data, str):
            argv.append(data)

    for obj in FUNCTIONS:
        AllFun.extend(FUNCTIONS[obj])

    if argv[0] in ("-h", "--help", "help", "/h", "/?"):
        if len(argv) > 1:
            if argv[1] in AllFun:
                help_function(argv[1])
            else:
                prt("Unknown function: ").fg_red(argv[1])
        else:
            help_function()
        return
    elif argv[0] in ("-l", "--list", "list"):
        list_functions()
        return

    funs = parse_cmd_argv(argv)
    try:
        for fun, args in funs:
            call_function(fun, args)
    except Exception as e:
        print(f"{fun}: {str(e)}")
        help_function(fun)


if __name__ == "__main__":
    cli()



import time
from typing import Union, Iterable
import sys
class ProgressBar:
    """ ## 进度条 
    未单开线程，如需及时刷新：update(0)"""
    CHARS = " ▏▎▍▌▋▊▉█" # 其实是半宽字符
    INVLEN = 1/len(CHARS)
    CENSUS = 2 # 统计speed的时间s
    FORMAT={
        'all': "{percentage}|{bar}| {progress} [{time}, {speed}]",
        'percentage': "{percentage:5.1f}%",
        'progress': "{processed:d}/{total:d}",
        'time': "{elapsed:.3f}<{remaining:.3f}",
        'speed': "{speed:.3f}{unit}/s"
    }
    def __init__(self, 
        iterator:Union[Iterable, None]=None, 
        total=0, 
        format="", 
        fmt_dict = None, 
        row=-1, col=-1, 
        ncols=10, 
        start=1, 
        interval=0.2, 
        unit='') -> None:
        """
        Args:
        - 0: iterator
        - format (str): 进度条格式 
          - progress
          - current/total, speed 
          - percent
          - elapsed, ramaining
          - 0[key]
        - row, col 位置
        - total 
        - color
        - ncols

        """
        self.format = format
        print(format)
        if total <= 0 and not isinstance(iterator, Iterable):
            total = -1
        self.num_total = total if total>0 else len(iterator)
        self.num_processed = 0
        self.ncols = ncols
        self.progress_bar = " " * self.ncols
        self.fmt_dict = fmt_dict if isinstance(fmt_dict, dict) else {}
        self.unit = unit
        self.speed = 0.0
        self.percentage = 0.0
        self.time_elapsed = 0.0
        self.time_remaining = 0.0
        self.time_total = 0.0
        self.time_start = 0.0
        self.time_last_draw = 0.0
        self.census_time = [0.0]
        self.census_num = [] # 最近两秒的数据池
        self.paused_start = 0.0
        self.paused_end = 0.0
        self.paused_time = 0.0
        self.__max_len = 0
        if start:
            self.start()


    def update(self, num):
        t = time.perf_counter()
        self.num_processed += num
        self.percentage = self.num_processed / self.num_total
        if self.census_num and t - self.census_time[0] > self.CENSUS:
            self.census_time.pop(0)
            self.census_num.pop(0)
        self.census_time.append(t)
        self.census_num.append(num)
        self.speed = sum(self.census_num)/(self.census_time[-1] - self.census_time[0])
        self.time_elapsed = t - self.time_start
        self.time_remaining = (self.num_total - self.num_processed) / self.speed # 使用CENSUS内的瞬时速度
        self.time_total = self.time_elapsed + self.time_remaining
        # self.draw()
    
    def draw(self):
        self.time_last_draw = time.perf_counter()
        n = self.percentage * self.ncols
        intN = int(n)
        remN = n - intN
        # print(n,remN,int(remN//self.INVLEN),self.CHARS[int(remN//self.INVLEN)])
        self.progress_bar = self.CHARS[-1] * intN + self.CHARS[int(remN//self.INVLEN)]
        self.progress_bar += ' '*(self.ncols - intN - 1)
        string = self.format.format(
            percentage=self.percentage*100, 
            bar=self.progress_bar, 
            processed = self.num_processed,
            total=self.num_total,
            elapsed = self.time_elapsed,
            remaining=self.time_remaining,
            speed = self.speed,
            unit=self.unit,
            **self.fmt_dict
        )
        l = len(string)
        if l >= self.__max_len:
            self.__max_len = l
        else:
            string += ' '*(self.__max_len-l)
        # sys.stdout.write('\r'+string)

    def start(self):
        if not self.time_start:
            self.time_start = time.perf_counter()
            self.census_time = [self.time_start]
            self.census_num = []
            self.draw()

    def pause(self):
        self.paused_start = time.perf_counter()

    def goon(self):
        self.paused_end = time.perf_counter()
        self.census_time = [self.paused_end]
        self.census_num = []
        self.paused_time += self.paused_end - self.paused_start



import re
import os
import sys
import time
import io
import cv2
import ctypes
import requests
from typing import Union
from PIL import Image
from unicodedata import east_asian_width
from platform import system as getOS
from threading import Thread
from pyaudio import PyAudio
from pydub import AudioSegment

__os = getOS()
IS_WIN = IS_LINUX = IS_MAC = False
if __os == "Windows":
    IS_WIN = True
elif __os == "Linux":
    IS_LINUX = True
elif __os == "Darwin":
    IS_MAC = True

# ------------------------------字符类处理函数---------------------------------
TabWidth = 4
Tab = " " * TabWidth

CHAR_WIDTH = {
    '—': 1
}

def getCharWidth(c: str):
    """返回字符宽度
    F W A ：全宽，Na、H：半宽，N：0
    """
    if c in CHAR_WIDTH:
        return CHAR_WIDTH[c]
    w = east_asian_width(c)
    if w == "N":
        # \t 应该返回多少宽度？
        return 0
    return 2 if w in ("F", "W", "A") else 1


def getStringWidth(s: str):
    """返回字符串去除CSI转义序列、\n、\t后的显示长度"""
    raw = re.sub(r"\033\[[\d;\?]*[a-zA-Z]", "", s)  # 去除csi转义序列
    return sum(getCharWidth(c) for c in raw)

def getEscapeString(s: str):
    """将一些不可见的控制字符转为可见的转义字符，包括空格32之前的和127 Delete（Oct：177）"""
    res = "".join([i.encode("unicode-escape").decode() if ord(i) < ord(" ") or i == "\177" else i for i in s])
    return res


def padString(s: str, width: int, mode=-1, fillchar=" "):
    """填充字符串s到宽度width （基于占位宽度）
    - mode: -1 左对齐右侧补充字符，0 居中对齐两边补充字符，1右对齐 左侧补充字符"""
    w = getStringWidth(s)
    if w >= width:
        return s
    width_fill_char = getStringWidth(fillchar)
    n = (width - w) // width_fill_char
    if mode == -1:
        s += fillchar * n
    elif mode == 0:
        h = n // 2
        s = f'{fillchar * h}{s}{fillchar * (n - h)}'
    elif mode == 1:
        s = fillchar * n + s
    else:
        raise ValueError(f"Parameter mode must be in -1,0,1 (got {mode}).")
    return s


# textwrap.wrap()的简化版，但是该方法不会解析转义序列，因此不采用
def splitLinesByWidth(s: str, width: int) -> list[str]:
    """按照显示宽度分割字符串，\\n 也会被分割，请不要包含 \\t 等字符，CSI转义序列会被保存但不计入宽度"""
    res, csi = [], []  # 结果，转义序列位置
    line, lwidth, i = [], 0, 0
    for match in re.finditer(r"\033\[[\d;\?]*[a-zA-Z]", s):
        csi.append(match.span())
    while i < len(s):
        chr = s[i]
        if csi and csi[0][0] == i:
            i = csi[0][1]
            line.append(s[csi[0][0] : csi[0][1]])
            csi.pop(0)
            continue
        if chr != "\n":
            line.append(chr)
            lwidth += getCharWidth(chr)
        if lwidth > width: # 最后一个字符是双宽字符导致宽度溢出时，使其换行
            last = line.pop()
            res.append(''.join(line))
            line = [last]
            lwidth = getCharWidth(last)
        elif lwidth == width or chr == "\n":
            res.append(''.join(line))
            line = []
            lwidth = 0
        i += 1
    if line:
        res.append(''.join(line))
    return res


# ------------------------------颜色类处理函数---------------------------------

RGB = Union[list[int], tuple[int, int, int], str]


def hex2RGB(hex: str) -> tuple[int, int, int]:
    """hex color to RGB color"""
    if hex[0] == "#":
        hex = hex[1:]
    hexes = []
    if len(hex) == 6:
        hexes = (hex[:2], hex[2:4], hex[4:])
    elif len(hex) == 3:
        hexes = (hex[:1] * 2, hex[1:2] * 2, hex[2:] * 2)
    else:
        raise ValueError("Hex color should be like #F0F or #00FFFF")
    return (int(hexes[0], 16), int(hexes[1], 16), int(hexes[2], 16))


def genGradient(color_start: RGB, color_end: RGB, num: int):
    """生成两个RGB颜色之间的渐变色列表
    - color_start: 起始颜色，格式为 (r, g, b)
    - color_end: 结束颜色，格式为 (r, g, b)
    - num: 总共要生成的渐变色数量（包括起始和结束颜色）

    Returns:
        包含起始颜色、渐变色和结束颜色的列表"""
    num -= 1
    if isinstance(color_start, str):
        color_start = hex2RGB(color_start)
    if isinstance(color_end, str):
        color_end = hex2RGB(color_end)
    step = [(color_end[i] - color_start[i]) / num for i in range(3)]
    gradient = [(int(color_start[0]+step[0]*i),
                int(color_start[1]+step[1]*i),
                int(color_start[2]+step[2]*i)) for i in range(num+1)]
    return gradient

def toImage(img_path: Union[str, Image.Image]):
    try:
        if isinstance(img_path,str):
            # 使用Path的请先str()转为字符串
            if os.path.exists(img_path):
                img = Image.open(img_path)
            elif img_path.startswith("http"):
                res = requests.get(img_path)
                img = Image.open(io.BytesIO(res.content))
            else:
                raise ValueError(f'Not a vaild path or url: {img_path}.')
        elif isinstance(img_path, Image.Image):
            img = img_path
        else:
            raise TypeError("Invalid type!")
    except Exception as e:
        raise ValueError(f"Parameter img_path({img_path}) is not "
                         "a valid image path or instance of Image.")
    return img

def getIMG(img_path: Union[str, Image.Image], height:int, width:int, resample=1):
    img = toImage(img_path)
    if img.mode != "RGB":
        img = img.convert("RGB")
    # 计算缩放比例
    img_width, img_height = img.size
    ratio_width = width / img_width
    ratio_height = height / img_height
    ratio = min(ratio_width, ratio_height)
    if ratio == 1:
        return img
    new_width = int(img_width * ratio)
    new_height = int(img_height * ratio)
    if new_height % 2:
        new_height += 1
    # 缩放图片
    img = img.resize((new_width, new_height), resample)
    return img

def readVideo2Image(video_path, max_frames=None):
    cap = cv2.VideoCapture(video_path)
    if not cap.isOpened():
        raise ValueError(f"Failed to open {video_path}")
    
    fps = cap.get(cv2.CAP_PROP_FPS)
    frame_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    
    frame_index = 0
    while True:
        ret, frame = cap.read()
        if not ret or (max_frames is not None and frame_num >= max_frames):
            break
        rgb_frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
        pil_image = Image.fromarray(rgb_frame)
        yield (pil_image, frame_index, frame_count, fps)
        frame_index += 1
    cap.release()


# ------------------------------时间类处理函数---------------------------------

PRECISE = 0.001

def sleepPrecise(seconds: float):
    start = time.perf_counter()
    n = seconds // PRECISE - 1 # -1 是因为时钟中断，实际暂停时间更久
    if n > 0:
        time.sleep(n * PRECISE)
    while time.perf_counter() - start < seconds:
        pass

# 循环等待的时间小于 2*PRECISE
def __win_sleepPrecise(seconds: float):
    start = time.perf_counter()
    n = seconds // PRECISE - 1
    if n > 0:
        winmm.timeBeginPeriod(1)
        time.sleep(n * PRECISE)
        winmm.timeEndPeriod(1)
    while time.perf_counter() - start < seconds:
        pass


if IS_WIN:
    winmm = ctypes.WinDLL('winmm')
    sleepPrecise = __win_sleepPrecise


# 通过耗时测试性能（本身也耗时，自测耗1μs左右）
class Timer:
    def __init__(self) -> None:
        self.t1 = time.perf_counter()
        self.t2 = 0
        self.span = 0

    def __enter__(self):
        self.t1 = time.perf_counter()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.t2 = time.perf_counter()
        self.span = self.t2 - self.t1
        print(f"{self.span:.9f}")

    def update(self):
        """ 获取与上一次update之间的时间间隔 """
        self.t2 = time.perf_counter()
        self.span = self.t2 - self.t1
        self.t1 = self.t2
        return self.span

TIMER = Timer()


class FrameTimer:
    def __init__(self, num: int=-1, spf: Union[int, float]=0, iterator=None) -> None:
        """ ## 帧计时器 (确保每一帧使用准确间隔)
        - num: 总帧数
        - spf: 每帧用时，等价于 1s / fps。未指定时请通过frameTime(seconds)指定当前帧应该耗费的时间
        - iterator: 可选，迭代器 或 可索引对象，长度必须>=num
        
        for i, v in FrameTimer:
            i = 当前帧数，从0到num-1
            v = None，或iterator迭代到相同次数的值，或对应下标的值"""
        self.__spf = spf
        self.__num = num if num>0 else 0x7fffffff
        self.__cur = 0
        self.__t0 = 0
        self.__total_spent = 0
        self.__last_frame_start = 0
        self.__iterator = iterator
        self.frame_spent = 0
        if self.__iterator == None:
            pass
        elif hasattr(self.__iterator, '__next__'):
            self.__getiter = lambda: self.__iterator.__next__()
        elif hasattr(self.__iterator, '__getitem__'):
            self.__getiter = lambda: self.__iterator[self.__cur - 1]

    def __iter__(self):
        return self
    
    def __next__(self):
        if self.__cur == 0:
            self.__t0 = time.perf_counter()
            self.__last_frame_start = self.__t0
            self.__cur += 1
            return 0, self.__getiter()
        if self.__cur >= self.__num:
            raise StopIteration
        i = self.__cur
        self.__cur += 1
        now = time.perf_counter()
        if self.__spf:
            self.__total_spent = now - self.__t0
            t = i * self.__spf - self.__total_spent
        else:
            self.frame_spent = now - self.__last_frame_start
            t = self.__frametime - self.frame_spent
        if t > 0: sleepPrecise(t)
        self.__last_frame_start = time.perf_counter()
        return i, self.__getiter()
    
    def __getiter(self):
        return None
    
    def frameTime(self, seconds: float):
        self.__frametime = seconds

    def stop(self):
        # Stop iteration (raise StopIteration at next iteration)
        self.__num = self.__cur

class Audio:
    def __init__(self, audio_path) -> None:
        audio = AudioSegment.from_file(audio_path)
        self.p = PyAudio()
        format = self.p.get_format_from_width(audio.sample_width)
        self.stream = self.p.open(
            format=format,
            channels=audio.channels,
            rate=audio.frame_rate,
            output=True,
        )
        self.__raw_data = audio.raw_data

    def __play(self, close=True):
        self.stream.write(self.__raw_data)
        if close:
            self.close()

    def play(self, close=True):
        Thread(target=self.__play, args=(close,), daemon=True).start()
    
    def close(self):
        self.stream.close()
        self.p.terminate()
